/*
 * FILE: gx.c
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this program handles all the AUX graphics library calls 
 */

#include <stdio.h> 
#include <math.h>
#include <GL/aux.h>
#include "view4d.h"

#define PERSP_4D_DIST		4
#define DEFAULT_VX 		50.
#define DEFAULT_VY 		50.
#define DEFAULT_VZ		50.
#define DEFAULT_VW		-50.
#define DEL_THETA_XW		2
#define DEL_THETA_XZ		2
#define DEL_THETA_XY		2
#define DEL_THETA_YZ		2
#define DEL_THETA_YW		2
#define DEL_THETA_ZW		2
#define SCALE_X			1.05
#define SCALE_Y			1.05
#define SCALE_Z			1.05
#define SCALE_W			1.05
#define THETA_DELTA		.05
#define Z_DELTA			2.0
#define ZOOM_DELTA		.01
#define XW			0
#define YW 			1
#define ZW			2
#define XY_AND_ZW		3
#define XW_AND_YZ		4
#define XZ_AND_YW		5
#define MANUAL			1
#define AUTOMATIC		2
#define NOTHING			0
#define SPIN			1
#define WIRE			0
#define SCAN_CONVERT		1	


#define INIT_WIDTH              512
#define INIT_HEIGHT             512
#define INIT_X			100
#define INIT_Y			100

/********** external variable definitions ****************/
struct polytope_struct *polytope;
GLint ntopes, draw_mode = WIRE;
GLdouble vx, vy, vz, vw;
GLint blackvec[3] = {0, 0, 0};
GLint whitevec[3] = {255, 255, 255};
GLdouble eyevec[3];
GLdouble org[3] = {0.0, 0.0, 0.0};
GLdouble idmat[4][4] = 
{
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,  
	0.0, 0.0, 0.0, 1.0 
};
GLdouble xyzwm[4][4], xzywm[4][4], xwm[4][4], xwyzm[4][4], ywm[4][4], zwm[4][4], rm[4][4]; 
GLint menu;
/**********************************************************/

/*********** function declarations *********************/
extern void go_graphics();
static void initialize();
static void get_inputs();
static void draw_polytopes();
static void draw_origin();
static void rotate_and_project_vertices_to_3D();
extern GLshort convert_to_hex();
extern void convert_to_ncube_coords();
extern void scale_to_screen_coords();
extern void sort_edge_table();
extern void read_polytopes();
extern void do_scan_convert();
static void rotate4D(int);
static void init_matrices();
static void init_rm_matrix();
static void sort_faces();
static void sort_vertices();
static void compute_centroid();
extern void compute_volume_normals();
extern void compute_surface_normals();
extern void compute_gouraud_4D_normals();
extern void compute_gouraud_3D_normals();
extern void depth_q();
extern void print_matrix();
extern void dump_frame_24bits();
extern void dump_frame_32bits();
extern void mark_time();
extern void clip3D();
extern void clear_image();
extern void sload();




static void Reshape(GLint width, GLint height)
{

    glViewport(0, 0, width, height);
    width_of_window = width;
    height_of_window = height;
}

/**********************************************************/
/* FUNCTION: Display_Func()			  	  */
/**********************************************************/
static void
Display_Func()
{
     draw_polytopes(draw_mode);

     glFlush();
}

/**********************************************************/
/* FUNCTION: draw_polytopes()			  	  */
/* PURPOSE: this function draws all the polytopes that are */
/*         defined. It uses the rotated normals to remove */
/* 	   backfaces and compute gouraud shadings.  	  */
/*	   NOTE: no gouraud shading takes place between   */
/*	   faces the are >= 90 degrees apart.		  */
/**********************************************************/
static void
draw_polytopes(type)
	GLint type;
{
	GLint fvtx, ipoly = 0, npols, jface, edge_num, iedge, jvtx, ivtx;
	struct point a,b;
	GLint edge_color[3];
	FILE *fp;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();           /* OGL addition */
	gluPerspective(90., 1.0, n, f);
	gluLookAt(vx,  vy, vz, 0., 0., 0., 0., 0., 1.);


	rotate_and_project_vertices_to_3D();

	edge_color[RED_IND] = O_DIFFUSE[0][RED_IND] * 255;
	edge_color[GRN_IND] = O_DIFFUSE[0][GRN_IND] * 255;
	edge_color[BLU_IND] = O_DIFFUSE[0][BLU_IND] * 255;

	if (type == SCAN_CONVERT) {
		
		clear_image(0., 0.,  0.);
		glDrawBuffer(GL_FRONT);
	        glClear(GL_COLOR_BUFFER_BIT);

		glFlush();

		sort_faces();
		sort_vertices();
		depth_q();

		if (polytope->num_cells > 1) {
			compute_volume_normals();
		        compute_gouraud_4D_normals();
		}
		else {
			compute_surface_normals();
	                compute_gouraud_3D_normals();
		}
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
	        fprintf(dfd, " compute_volume_normals done\n");
		fclose(dfd);
#endif

		(void) initialize_edge_table();
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
        	fprintf(dfd, " initialize_edge_table done\n");
		fclose(dfd);
#endif
 
        	(void) convert_to_ncube_coords();
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
        	fprintf(dfd, " convert_to_ncube_coords done\n");
		fclose(dfd);
#endif

		(void) clip3D();
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
		fprintf(dfd, " clipping done\n");
		fclose(dfd);
#endif

	        (void) scale_to_screen_coords();
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
        	fprintf(dfd, " scale_to_screen_coords done\n");
		mark_time("scan_conversion start");
#endif

	        /* add ALL faces to ET */
		for (jface=0;jface<polytope->num_faces;jface++) {

		        if (polytope->face[jface].visible) {
#ifdef DEBUG_SCAN
			   dfd = fopen(DEBUGFILE, "a");
		           fprintf(dfd, " face %d is visible\n", jface);
			   fclose(dfd);
#endif
			   add_face_to_ET(jface, jface);
			}
			ipoly++;
		
	        }
	
		npols = ipoly - 1;
		(void) sort_edge_table();
 
#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
		fprintf(dfd, " sort_edge_table done\n");
		(void) print_ET();
		fclose(dfd);
#endif 

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();           /* OGL addition */
		gluOrtho2D(0, width_of_window, 0, height_of_window);
 
        	(void) do_scan_convert(polytope->num_faces);

#ifdef DEBUG_SCAN
		dfd = fopen(DEBUGFILE, "a");
		mark_time("scan_conversion end");
		fclose(dfd);
#endif 

		if (archive_it) {
		    FILE *archive_fd;
  		    if ((archive_fd = fopen(archive_name,"w")) == NULL) {
                          printf("Unable to open %s...Exiting\n", archive_name);
                	  exit(1);
        	    }	
		    dump_frame_32bits(archive_fd);
		    fclose(dfd);
	            fclose(archive_fd);
		    printf(
			"image archived to %s -- use preview to view\n", 
				archive_name);
		}

		if (polytope->clipped) {
			polytope->clipped = GL_FALSE;
			read_polytopes(input_file);
		}
	        draw_mode = WIRE;
	} 
	else {

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();          
	gluPerspective(90., 1.0, n, f);
	gluLookAt(vx,  vy, vz, 0., 0., 0., 0., 0., 1.);
	
          glClear(GL_COLOR_BUFFER_BIT);
          glColor3ub(edge_color[RED_IND],
                        edge_color[GRN_IND], edge_color[BLU_IND]);
          for (iedge=0;iedge<polytope->num_edges;iedge++) {
                glBegin(GL_LINES);
                  glVertex3dv(polytope->v3D[polytope->edge[iedge].a].p);
                  glVertex3dv(polytope->v3D[polytope->edge[iedge].b].p);
                glEnd();
          }

        }  

}

/**********************************************************/
/* FUNCTION: rotate_and_project_vertices_to_3D()  	  */
/* PURPOSE: this function rotate ALL 4D vertices   by     */
/* 	the current rotation matrix and then projects     */
/* 	them down to 3D					  */
/**********************************************************/
static void
rotate_and_project_vertices_to_3D()
{
    GLint ivert;
    GLdouble denom;

    for (ivert = 0; ivert < polytope->num_vertices; ivert++) {

	/* rotate by 4D rotation matrices */
        polytope->rot_vertex[ivert].p[X_COORD] = 
		polytope->vertex[ivert].p[X_COORD]*rm[0][X_COORD] +
		polytope->vertex[ivert].p[Y_COORD]*rm[1][X_COORD] +
		polytope->vertex[ivert].p[Z_COORD]*rm[2][X_COORD] +
		polytope->vertex[ivert].p[W_COORD]*rm[3][X_COORD];
 	polytope->rot_vertex[ivert].p[Y_COORD] = 
		polytope->vertex[ivert].p[X_COORD]*rm[0][Y_COORD] +
		polytope->vertex[ivert].p[Y_COORD]*rm[1][Y_COORD] +
		polytope->vertex[ivert].p[Z_COORD]*rm[2][Y_COORD] +
		polytope->vertex[ivert].p[W_COORD]*rm[3][Y_COORD];
	polytope->rot_vertex[ivert].p[Z_COORD] = 
		polytope->vertex[ivert].p[X_COORD]*rm[0][Z_COORD] +
		polytope->vertex[ivert].p[Y_COORD]*rm[1][Z_COORD] +
		polytope->vertex[ivert].p[Z_COORD]*rm[2][Z_COORD] +
		polytope->vertex[ivert].p[W_COORD]*rm[3][Z_COORD];
	polytope->rot_vertex[ivert].p[W_COORD] = 
		polytope->vertex[ivert].p[X_COORD]*rm[0][W_COORD] +
		polytope->vertex[ivert].p[Y_COORD]*rm[1][W_COORD] +
		polytope->vertex[ivert].p[Z_COORD]*rm[2][W_COORD] +
		polytope->vertex[ivert].p[W_COORD]*rm[3][W_COORD];

	/* transform down to 3D */
	denom = polytope->rot_vertex[ivert].p[W_COORD] + PERSP_4D_DIST;
 	polytope->v3D[ivert].p[X_COORD] = 
		polytope->rot_vertex[ivert].p[X_COORD]/denom;
 	polytope->v3D[ivert].p[Y_COORD] = 
		polytope->rot_vertex[ivert].p[Y_COORD]/denom;
 	polytope->v3D[ivert].p[Z_COORD] = 
		polytope->rot_vertex[ivert].p[Z_COORD]/denom;

	/* now scale and move to polytope center */
	polytope->v3D[ivert].p[X_COORD] = 
	    polytope->scale_fac*polytope->v3D[ivert].p[X_COORD] 
		+ polytope->origin.p[X_COORD];
	polytope->v3D[ivert].p[Y_COORD] = 
	    polytope->scale_fac*polytope->v3D[ivert].p[Y_COORD] 
		+ polytope->origin.p[Y_COORD];
	polytope->v3D[ivert].p[Z_COORD] = 
	   polytope->scale_fac*polytope->v3D[ivert].p[Z_COORD] 
		+ polytope->origin.p[Z_COORD];
  	polytope->v3D[ivert].p[W_COORD] = 1.0;

      }
	
}

/**********************************************************/
/* FUNCTION: draw_origin()			  	  */
/* PURPOSE: this function draws the x, y, and z coord.    */
/*		axes with a solid, dashed, and dotted     */
/*		white line, respectively to help the      */
/*		viewer gain perspective to the image.     */
/***********************************************************/
static void
draw_origin(length)
	GLdouble length;
{
	GLdouble or_vx[3], or_vy[3], or_vz[3];

	glColor3ub(whitevec[RED_IND], whitevec[GRN_IND], whitevec[BLU_IND]);
	
	or_vx[X_COORD] = length;
	or_vx[Y_COORD] = 0.0;
	or_vx[Z_COORD] = 0.0;
	
	glBegin(GL_LINES);
		glVertex3dv(org);
		glVertex3dv(or_vx);
	glEnd();	

	or_vy[Y_COORD] = length;
	or_vy[X_COORD] = 0.0;
	or_vy[Z_COORD] = 0.0;
	
	glLineStipple(1, 0xF0F0);
	glBegin(GL_LINES);
		glVertex3dv(org);
		glVertex3dv(or_vy);
	glEnd();	
	
	glLineStipple(1, 0x8888);

	or_vz[Z_COORD] = length;
	or_vz[Y_COORD] = 0.0;
	or_vz[X_COORD] = 0.0;
	
	glBegin(GL_LINES);
		glVertex3dv(org);
		glVertex3dv(or_vz);
	glEnd();	
	glLineStipple(1, 0x0);
}
		

/**********************************************************/
/* FUNCTION: rotate4D(int, int)			  	  */
/* PURPOSE: this function uses the graphics hardware to	  */
/* 		rotate the vertices in 4D		  */
/**********************************************************/
static void
rotate4D(axes)
	GLint axes;
{
	switch (axes) {
		case XY_AND_ZW:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(xyzwm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;
		case XW_AND_YZ:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(xwyzm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;	
		case XZ_AND_YW:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(xzywm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;	
		case XW:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(xwm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;
		case YW:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(ywm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;		
		case ZW:
			glPushMatrix();
			glLoadMatrixd(rm);
			glMultMatrixd(zwm);
			glGetDoublev(GL_PROJECTION_MATRIX, rm);
			glPopMatrix();
			break;	
	}
}

/**********************************************************/
/* FUNCTION: init_matrices()			  	  */
/* PURPOSE: this function initializes the rotation 	  */
/* 			matrices 			  */
/**********************************************************/
static void
init_matrices()
{
    GLint iz, izz;

    for (iz =0;iz<4;iz++) 
	for (izz=0;izz<4;izz++) {
	    if (iz != izz) {
		xyzwm[iz][izz] = 0.0;
		xzywm[iz][izz] = 0.0;
		xwm[iz][izz] = 0.0;
		xwyzm[iz][izz] = 0.0;
		ywm[iz][izz] = 0.0;
		zwm[iz][izz] = 0.0;
	    } else {
		xyzwm[iz][izz] = 1.0;
		xzywm[iz][izz] = 1.0;
		xwm[iz][izz] = 1.0;
		xwyzm[iz][izz] = 1.0;
		ywm[iz][izz] = 1.0;
		zwm[iz][izz] = 1.0;
	    }
	}
  	     
    xyzwm[X_COORD][X_COORD] = xyzwm[Y_COORD][Y_COORD] = 
				cos(DEL_THETA_XY/180.*M_PI);
    xyzwm[Y_COORD][X_COORD] = sin(DEL_THETA_XY/180.*M_PI);
    xyzwm[X_COORD][Y_COORD] = -xyzwm[Y_COORD][X_COORD];
    xyzwm[Z_COORD][Z_COORD] = xyzwm[W_COORD][W_COORD] = 
				cos(DEL_THETA_ZW/180.*M_PI);
    xyzwm[W_COORD][Z_COORD] = sin(DEL_THETA_ZW/180.*M_PI);
    xyzwm[Z_COORD][W_COORD] = -xyzwm[W_COORD][Z_COORD];
    xzywm[X_COORD][X_COORD] = xzywm[Z_COORD][Z_COORD] = 
				cos(DEL_THETA_XZ/180.*M_PI);
    xzywm[Z_COORD][X_COORD] = sin(DEL_THETA_XZ/180.*M_PI);
    xzywm[X_COORD][Z_COORD] = -xzywm[Z_COORD][X_COORD];
    xzywm[Y_COORD][Y_COORD] = xzywm[W_COORD][W_COORD] = 
				cos(DEL_THETA_YW/180.*M_PI);
    xzywm[W_COORD][Y_COORD] = sin(DEL_THETA_YW/180.*M_PI);
    xzywm[Y_COORD][W_COORD] = -xzywm[W_COORD][Y_COORD];
    xwyzm[X_COORD][X_COORD] = xwyzm[W_COORD][W_COORD] = 
				cos(DEL_THETA_XW/180.*M_PI);
    xwyzm[W_COORD][X_COORD] = sin(DEL_THETA_XW/180.*M_PI);
    xwyzm[X_COORD][W_COORD] = -xwyzm[W_COORD][X_COORD];
    xwyzm[Y_COORD][Y_COORD] = xwyzm[Z_COORD][Z_COORD] = 
				cos(DEL_THETA_YZ/180.*M_PI);
    xwyzm[Z_COORD][Y_COORD] = sin(DEL_THETA_YZ/180.*M_PI);
    xwyzm[Y_COORD][Z_COORD] = -xwyzm[Z_COORD][Y_COORD];
    xwm[X_COORD][X_COORD] = xwm[W_COORD][W_COORD] = 
				cos(DEL_THETA_XW/180.*M_PI);
    xwm[W_COORD][X_COORD] = sin(DEL_THETA_XW/180.*M_PI);
    xwm[X_COORD][W_COORD] = -xwm[W_COORD][X_COORD];
    ywm[Y_COORD][Y_COORD] = ywm[W_COORD][W_COORD] = 
				cos(DEL_THETA_YW/180.*M_PI);
    ywm[W_COORD][Y_COORD] = sin(DEL_THETA_YW/180.*M_PI);
    ywm[Y_COORD][W_COORD] = -ywm[W_COORD][Y_COORD];
    zwm[Z_COORD][Z_COORD] = zwm[W_COORD][W_COORD] = 
				cos(DEL_THETA_ZW/180.*M_PI);
    zwm[W_COORD][Z_COORD] = sin(DEL_THETA_ZW/180.*M_PI);
    zwm[Z_COORD][W_COORD] = -zwm[W_COORD][Z_COORD];


}

static void
init_rm_matrix()
{
    GLint iz, izz;
   
    for (iz =0;iz<4;iz++) 
       for (izz =0;izz<4;izz++) 
	    if (iz != izz)  rm[iz][izz] = 0.0;
	    else rm[iz][izz] = 1.0;
}

void
print_matrix(name, mat)
	char *name;
	GLdouble mat[4][4];
{
	GLint ip;

	dfd = fopen(DEBUGFILE, "a");
	fprintf(dfd, "\n\tMATRIX %s\n\n", name);
	for (ip = 0;ip < 4; ip++) 
	    fprintf(dfd, "%10.6f %10.6f %10.6f %10.6f\n",
	       mat[ip][0], mat[ip][1], mat[ip][2], mat[ip][3]);
	fclose(dfd);
}

/**********************************************************/
/* FUNCTION: sort_faces()			  	  */
/* PURPOSE: this function sort all faces of the polytope  */
/* 			by distance			  */
/**********************************************************/
static void
sort_faces()
{
	GLint iface, jface, ivtx, temp;
	GLdouble x_comps, y_comps, z_comps, w_comps;

	compute_centroid();

	for (iface = 0; iface < polytope->num_faces; iface++) {
		
		/* compute center of each face */
		x_comps = 0.0;y_comps = 0.0;z_comps = 0.0;w_comps = 0.0;
		for (ivtx = 0;ivtx<polytope->face[iface].num_vertices;
			ivtx++) {
		     x_comps = x_comps + polytope->rot_vertex[
			polytope->face[iface].vertex[ivtx]].p[X_COORD];
		     y_comps = y_comps + polytope->rot_vertex[
			polytope->face[iface].vertex[ivtx]].p[Y_COORD];
		     z_comps = z_comps + polytope->rot_vertex[
			polytope->face[iface].vertex[ivtx]].p[Z_COORD]; 
		     w_comps = w_comps + polytope->rot_vertex[
			polytope->face[iface].vertex[ivtx]].p[W_COORD];
		}


		x_comps = x_comps/
			((GLdouble) polytope->face[iface].num_vertices);
		y_comps = y_comps/
			((GLdouble) polytope->face[iface].num_vertices);
		z_comps = z_comps/
			((GLdouble) polytope->face[iface].num_vertices);
		w_comps = w_comps/
			((GLdouble) polytope->face[iface].num_vertices);

		polytope->face[iface].depth = 
	         DIST(x_comps, vx, y_comps, vy, z_comps, vz, w_comps, vw);
	
	       	polytope->ordered_faces[iface] = iface;

	    }

	    /* now sort faces in non-decreasing order */
            for (iface = polytope->num_faces-1; iface>0; iface--) {
	
		for (jface = 0;jface < iface ; jface ++) {
		    if (FP_LT(
	polytope->face[polytope->ordered_faces[jface]].depth,
	polytope->face[polytope->ordered_faces[iface]].depth)) {
			temp = polytope->ordered_faces[iface];
			polytope->ordered_faces[iface] = 
				polytope->ordered_faces[jface];
			polytope->ordered_faces[jface] = temp;
		    }
		}
	     }

		    
#ifdef DEBUG_SCAN
	    dfd = fopen(DEBUGFILE, "a");
	    fprintf(dfd, "FACES LISTED FROM FAR TO NEAR:\n");
	    for (iface = 0; iface < polytope->num_faces; iface++)
		fprintf(dfd, "face %d dist: %lf \n", 
			polytope->ordered_faces[iface], 
			polytope->face[
		polytope->ordered_faces[iface]].depth);  
	    fclose(dfd);
#endif

}
	
/**********************************************************/
/* FUNCTION: compute_centroid()			  	  */
/* PURPOSE: this function computes centroid of polytope   */
/**********************************************************/
static void
compute_centroid()
{
	GLint ivtx;
	GLdouble x_comps, y_comps, z_comps, w_comps;
	
	x_comps = 0.0;y_comps = 0.0;z_comps = 0.0;w_comps = 0.0;
	for (ivtx = 0;ivtx<polytope->num_vertices;ivtx++) {
	    x_comps = x_comps + polytope->rot_vertex[ivtx].p[X_COORD];
            y_comps = y_comps + polytope->rot_vertex[ivtx].p[Y_COORD];
            z_comps = z_comps + polytope->rot_vertex[ivtx].p[Z_COORD]; 
            w_comps = w_comps + polytope->rot_vertex[ivtx].p[W_COORD];
	}

	polytope->centroid.p[X_COORD] = vx;
	polytope->centroid.p[Y_COORD] = vy;
	polytope->centroid.p[Z_COORD] = vz;
        polytope->centroid.p[W_COORD] = vw;

}

/**********************************************************/
/* FUNCTION: sort_vertices()			  	  */
/* PURPOSE: this function sort all vertices of polytope   */
/* 			by distance			  */
/**********************************************************/
static void
sort_vertices()
{
	GLint ivtx, jvtx, temp;
	GLdouble x_comps, y_comps, z_comps, w_comps;


	    for (ivtx = 0; ivtx < polytope->num_vertices; ivtx++) {
		
		/* compute dist of each vertex from viewer */
		polytope->vertex[ivtx].depth = DIST(
		  polytope->rot_vertex[ivtx].p[X_COORD], vx, 
		  polytope->rot_vertex[ivtx].p[Y_COORD], vy, 
		  polytope->rot_vertex[ivtx].p[Z_COORD], vz,
		  polytope->rot_vertex[ivtx].p[W_COORD], vw);

			
	        polytope->ordered_vertices[ivtx] = ivtx;

	    }

#ifdef DEBUG_SCAN
	    dfd = fopen(DEBUGFILE, "a");
	    for (ivtx = 0; ivtx < polytope->num_vertices; ivtx++)
		fprintf(dfd, "%d ",polytope->ordered_vertices[ivtx]);
	    fclose(dfd);
#endif

	    /* now sort vertices in non-increasing order */
            for (ivtx = polytope->num_vertices-1; ivtx>0; ivtx--) {
	
		for (jvtx = 0;jvtx < ivtx ; jvtx ++) {
		    if (FP_LT(
	polytope->vertex[polytope->ordered_vertices[jvtx]].depth,
        polytope->vertex[polytope->ordered_vertices[ivtx]].depth)) {
			temp = polytope->ordered_vertices[ivtx];
			polytope->ordered_vertices[ivtx] = 
	polytope->ordered_vertices[jvtx];
			polytope->ordered_vertices[jvtx] = temp;
		    }
	         }
	     }

#ifdef DEBUG_SCAN
	    dfd = fopen(DEBUGFILE, "a");
	    fprintf(dfd, "VERTICES LISTED FROM FAR TO NEAR:\n");
	    for (ivtx = 0; ivtx < polytope->num_vertices; ivtx++)
		fprintf(dfd, "vertex %d dist: %lf \n", 
			polytope->ordered_vertices[ivtx], 
			polytope->vertex[
		polytope->ordered_vertices[ivtx]].depth);  
	    fclose(dfd);
#endif


}

extern void
clear_image(r, g, b)
	GLfloat r, g, b;
{
	GLint x, y;

	for (x = 0; x < width_of_window; x++)
		for (y = 0; y < height_of_window; y++) {
			image[x][y] = r;
			image[x][y] = g;
			image[x][y] = b;
		}
}

static GLenum Key(int key, GLenum mask)
{
        GLdouble theta, mz, nvx, nvy;

    switch (key) {
      case AUX_o:
      case AUX_O:
                    vx = vx + vx*ZOOM_DELTA;
                    vy = vy + vy*ZOOM_DELTA;
                    vz = vz + vz*ZOOM_DELTA;
		    break;
      case AUX_i:
      case AUX_I:
                    vx = vx - vx*ZOOM_DELTA;
                    vy = vy - vy*ZOOM_DELTA;
                    vz = vz - vz*ZOOM_DELTA;
		    break;
      case AUX_U:
      case AUX_u:
                   vz = vz + Z_DELTA;
		   break;
      case AUX_D:
      case AUX_d:
                   vz = vz - Z_DELTA;
		   break;
      case AUX_Q:
      case AUX_q:
      case AUX_ESCAPE:
		    auxQuit();
		    break;
      case AUX_r:
      case AUX_R:
                    /* move in theta planes */
                    theta = -1*THETA_DELTA;
                    nvx = vx*cos(theta) - vy*sin(theta);
                    nvy = vy*cos(theta) + vx*sin(theta);
                    vx = nvx; vy = nvy;
		    break;
      case AUX_l:
      case AUX_L:
                    /* move in theta planes */
                    theta = THETA_DELTA;
                    nvx = vx*cos(theta) - vy*sin(theta);
                    nvy = vy*cos(theta) + vx*sin(theta);
                    vx = nvx; vy = nvy;
		    break;
      case AUX_S:
      case AUX_s:
		    draw_mode = SCAN_CONVERT;
		    break;
      case AUX_1:
              	    /* rotate  XY_AND_ZW */
                    rotate4D(XY_AND_ZW);
                    break;
      case AUX_2:
                    /* rotate XW_AND_YZ */
                    rotate4D(XW_AND_YZ);
                    break;
      case AUX_3:
                    /* rotate XZ_AND_YW */
                    rotate4D(XZ_AND_YW);
                    break;
      case AUX_X:
      case AUX_x:
                    /* rotate XW */
                    rotate4D(XW);
                    break;
      case AUX_Y:
      case AUX_y:
            	    /* rotate YW */
                    rotate4D(YW);
                    break;
      case AUX_Z:
      case AUX_z:
                    /* rotate ZW */
                    rotate4D(ZW);
                    break;
      case AUX_0: 	
		    /* reset */
    		    glLoadIdentity();
                    vx = DEFAULT_VX;
                    vy = DEFAULT_VY;
                    vz = DEFAULT_VZ;
                    vw = DEFAULT_VW;
                    init_rm_matrix();
                    break;
    }
    return GL_TRUE;
}

/**********************************************************/
/* FUNCTION: go_graphics()				  */
/* PURPOSE: this function controls the graphics routines  */
/**********************************************************/
void
go_graphics()
{
	(void) initialize();
	draw_polytopes(WIRE);
	auxMainLoop(Display_Func);

}

/**********************************************************/
/* FUNCTION: initialize()				  */
/* PURPOSE:   this sets up all the graphics settings and  */
/* 	devices.					  */
/**********************************************************/
void
initialize()
{
	GLenum type;
	GLint vp[4];
	GLdouble lookmat[4][4];

	/* initialize all graphics settings
	 * 	use 3 light sources
	 * 	  1) white at z = INFINITY
	 * 	  2) blue at y = 1000. 
	 * 	  3) red at x = 1000.
	 *	use GOURAUD shading
         *	turn zbuffer hidden surface removal ON
	 */

	auxInitPosition(INIT_X, INIT_Y, INIT_WIDTH, INIT_HEIGHT);
	width_of_window = INIT_WIDTH;
	height_of_window = INIT_HEIGHT;

	type = AUX_RGB | AUX_SINGLE | AUX_INDIRECT;
        auxInitDisplayMode(type);


	if (auxInitWindow("4D Polytope Viewer") == GL_FALSE) {
		printf("cannot InitWindow\n");
		auxQuit();
	}

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();           /* OGL addition */
	gluPerspective(90., 1.0, n, f);
	glGetDoublev(GL_PROJECTION_MATRIX, persp_mat);
	/* set up initial viewpoint */
	vx = DEFAULT_VX;
	vy = DEFAULT_VY;
	vz = DEFAULT_VZ;
	vw = DEFAULT_VW;
	glShadeModel(GL_FLAT);
	glDisable(GL_DEPTH_TEST);
	glClearColor(0., 0., 0., 0.);

	dfd = fopen(DEBUGFILE, "a");
        glGetIntegerv(GL_VIEWPORT, vp);
	l = vp[0];
	r = vp[0] + vp[2];
	b = vp[1];
	t = vp[1] + vp[3];
        fprintf(dfd, " viewport l: %hd r: %hd B: %hd t: %hd\n",
                       	l, r, b, t); 
        glGetIntegerv(GL_DEPTH_RANGE, vp);
        fprintf(dfd, " depth: n:%hd f:%hd \n", vp[0], vp[1]);
	fclose(dfd);

        init_matrices();
	init_rm_matrix();

    	auxKeyDownFunc(Key);
	auxReshapeFunc(Reshape);
     
}
