/*
 * FILE: view4d.c
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this program reads any number of 4D objects defined in an input file
 */

#include <stdio.h> 
#include <math.h>
#include <GL/aux.h>
#include "view4d.h"

/********** external variable definitions ****************/
struct polytope_struct *polytope;
GLboolean archive_it = GL_FALSE;
struct PT_struct *PT;
struct ET_element **ET;
struct ET_element *AET;
GLint r, l, t, b, n=3, f=10000;
GLdouble persp_mat[16];
char archive_name[MAX_FILE_NAME_LEN], input_file[MAX_FILE_NAME_LEN];
FILE *dfd;
GLint image[MAX_PIXELS][MAX_PIXELS];
GLint width_of_window, height_of_window;
/**********************************************************/

/*********** function declarations *********************/
extern void read_polytopes();
static void write_polytopes();
extern void transf_vtx_to_screen();
extern void convert_to_ncube_coords();
extern void scale_to_screen_coords();
extern void print_point();
extern void compute_gouraud_4D_normals();
extern void compute_gouraud_3D_normals();
static struct point vector_minus();
static GLboolean cell_contains_vtx();
static GLboolean face_contains_vtx();
extern void compute_surface_normals();
extern void compute_volume_normals();
extern void normalize();
extern double det();
extern void mark_time();
extern GLboolean cross();
/********************************************************/

static GLenum Args(int argc, char **argv)
{
 	if ((argc == 2) || (argc == 3)) {
		if (strcpy(input_file, argv[1]) == 0) {
			printf("input file name incorrect\n");
			return GL_FALSE;
		}
		read_polytopes(input_file);
		write_polytopes(input_file);
		if (argc == 3) {
		   if (strcpy(archive_name, argv[2]) == 0) {
				printf("archive file name incorrect\n");
				return GL_FALSE;
		   }
		   archive_it = GL_TRUE;
		}
	}
   	else {
               printf("Usage: %s <description-file> "
			"[image_archive_file]!!!!\n", argv[0]);
	       return GL_FALSE;
    	}       
    	return GL_TRUE;
}

void
main(argc, argv)
	int argc;
	char *argv[];
{
	if (Args(argc, argv) == GL_FALSE) {
        	auxQuit();
    	}

	polytope->clipped = GL_FALSE;
  
	(void) go_graphics();  
   
}


/********************************************************/
/* FUNCTION: read_polytopes()				*/
/* PURPOSE:  this function reads polytope decriptions   */
/*	     from an input file in the following format:*/
/*     LINE 1: # of polytopes				*/
/*     for each polytope:				*/
/*	  position of polytope center in x, y, z, w     */
/*	  scale factor					*/
/*        # of vertices			 		*/
/*          coords of each vertex         		*/
/*        # of edges					*/
/*	    starting and ending vertex of each edge 	*/
/*	  # of faces					*/
/*	  # of vertices in face and each vtx number 	*/
/********************************************************/
extern void
read_polytopes(in_name)
	char *in_name;
{

	GLint icell, ivert;
	GLint iedge, iface, ivf, ied, totpols;
	static GLboolean first_time = GL_TRUE;
	FILE *fd;
	

        if ((fd = fopen(in_name,"r")) == NULL) {
		printf("Unable to open %s...Exiting\n", in_name);
		exit(1);
	}
	/****** 
	*	get details of each polytope 
	*******/
	if (first_time) polytope = 
		(struct polytope_struct *) malloc(sizeof(struct polytope_struct));


	/* get scale factor for polytope */
	if (fscanf(fd,"%lf \n",&polytope->scale_fac) != 1) {
			printf(" ERROR reading scale_fac\n");
			exit(0);
	}

	/* get absolute position of origin of polytope */
	if (fscanf(fd,"%lf %lf %lf %lf\n",
			&polytope->origin.p[X_COORD], 
			&polytope->origin.p[Y_COORD],
			&polytope->origin.p[Z_COORD],
                        &polytope->origin.p[W_COORD]) 
			!= 4)  {
			    printf(" ERROR reading origin position\n");
			    exit(0);
	}

	/* get vertices in relative coordinates from origin */
	if (fscanf(fd,"%d \n",&polytope->num_vertices) != 1) {
			printf(" ERROR reading nverts\n");
			exit(0);
	}

	if (first_time) {
	   polytope->vertex = 
		(struct point *) malloc((polytope->num_vertices) * 
					sizeof(struct point));
	   polytope->rot_vertex = (struct point *) 
				malloc((MAX_VERTICES) * 
					sizeof(struct point));
	   polytope->v3D = (struct point *) 
				malloc((polytope->num_vertices) * 
					sizeof(struct point));
	   polytope->screen_vtx = (struct point *) 
				malloc((MAX_VERTICES) * 
					sizeof(struct point));
	   polytope->gouraud_norm = (struct point *) 
				malloc((MAX_VERTICES) * 
					sizeof(struct point));
	   polytope->ordered_vertices = (int *) 
				malloc((polytope->num_vertices) * 
					sizeof(int));
	}
		
	for (ivert = 0; ivert < polytope->num_vertices;ivert++)
	{
		if (fscanf(fd,"%lf %lf %lf %lf\n",
			  &polytope->vertex[ivert].p[X_COORD],
			  &polytope->vertex[ivert].p[Y_COORD],
			  &polytope->vertex[ivert].p[Z_COORD],
			  &polytope->vertex[ivert].p[W_COORD]) != 4) {
				printf(" ERROR reading vertices\n");
				exit(0);
		}
	}
	/************************/

	/* get edges of polytope */
	if (fscanf(fd,"%d \n",&polytope->num_edges) != 1) {
			printf(" ERROR reading nedges");
			exit(1);
	}

	if (first_time) polytope->edge = 
		(struct edge_struct *) malloc((MAX_EDGES)
		* sizeof(struct edge_struct));
		
	for (iedge = 0;iedge < polytope->num_edges;iedge++) {
			
		if (fscanf(fd,"%d %d \n",
			  &polytope->edge[iedge].a,
			  &polytope->edge[iedge].b) != 2) {
				printf(" ERROR reading edges\n");
				exit(1);
		}
	}
	/************************/

	/***** get faces of polytope *****/

	if (fscanf(fd,"%d \n",&polytope->num_faces) != 1) {
			printf(" ERROR reading nfaces");
			exit(1);
	}

	if (first_time) {
	   polytope->face = (struct face_struct *)
		malloc((polytope->num_faces) * sizeof(struct face_struct));

	   polytope->ordered_faces = (int *)
		malloc((polytope->num_faces) * sizeof(int));
	}
		
	for (iface = 0; iface < polytope->num_faces;iface++) {
		  
		  /* get vertices of each face */
		  if (fscanf(fd,"%d ",
		    &polytope->face[iface].num_vertices) != 1) {
			printf(" ERROR reading num_vertices for face\n");
			exit(1);
	           }

		   if (first_time) polytope->face[iface].vertex = 
			(int *) malloc((MAX_FACE_VERTICES) *
				sizeof(int));

		   for (ivf = 0;ivf < 
		     polytope->face[iface].num_vertices;ivf++) {
					
			 if (fscanf(fd,"%d ", 
			     &polytope->face[iface].vertex[ivf])!= 1) {
				printf(" ERROR reading face vertices\n");
				exit(1);
			    }
		    }

		   /* get edges of each face */
		  
		   polytope->face[iface].num_edges =
			 polytope->face[iface].num_vertices;

		   if (first_time) polytope->face[iface].edge = (int *)
			malloc((MAX_FACE_EDGES) * sizeof(int));

		   for (ied = 0;ied < 
		     polytope->face[iface].num_edges;ied++) {
					
			 if (fscanf(fd,"%d ", 
			     &polytope->face[iface].edge[ied])!= 1) {
				printf(" ERROR reading face edges\n");
				exit(1);
			    }
		    }

		    fscanf(fd," \n");

	}

	/********* get cells of the polytope *********/
        if (fscanf(fd,"%d\n",&polytope->num_cells) != 1) {
                        printf(" ERROR reading num_cells\n");
                        exit(1);
      	}
 
	if (first_time) polytope->cell = (struct cell_struct *)
            malloc((polytope->num_cells + 1) * sizeof(struct cell_struct));

        for (icell = 0;icell < polytope->num_cells;icell++) {

        	if (fscanf(fd,"%d\n",&polytope->cell[icell].num_faces) != 1) {
                        printf(" ERROR reading nfaces for cell %d\n", icell);
                        exit(1);
        	}
		if (first_time) polytope->cell[icell].face = 
			(struct face_struct *)
                       malloc((polytope->cell[icell].num_faces)
                       * sizeof(struct face_struct));

        	for (iface = 0; iface < polytope->cell[icell].num_faces; 
			iface++) {
 
                   if (fscanf(fd,"%d ",
			&polytope->cell[icell].face[iface].num_vertices)
                        != 1) {
                        printf(" ERROR reading nverts of face %d for cell %d\n",
				 iface, icell);
                        exit(1);
                   }
 
		   if (first_time) polytope->cell[icell].face[iface].vertex = 
	           (int *) malloc((polytope->cell[icell].face[iface].num_vertices)
                                        * sizeof(int));
 
                   for (ivert = 0;
		ivert < polytope->cell[icell].face[iface].num_vertices; ivert++) {
                        if (fscanf(fd,"%d ",
                          &polytope->cell[icell].face[iface].vertex[ivert]) != 1) {
                     printf(
		" ERROR reading face edge for face %d of cell %d\n", iface, icell);
                                exit(1);
                        }
                   }
                   fscanf(fd,"\n");
                }   
	}

	/* get coefficients */
	fscanf(fd,"%lf %lf %*s\n", &close_tsp, &far_tsp);
	fscanf(fd,"%lf %lf %*s\n", &close_amb, &far_amb);
	fscanf(fd,"%lf %lf %*s\n", &close_diffuse, &far_diffuse);
	fscanf(fd,"%lf %lf %*s\n", &close_spec, &far_spec);


	O_DIFFUSE[0][RED_IND] = 1.;
	O_DIFFUSE[0][GRN_IND] = 1.;
 	O_DIFFUSE[0][BLU_IND] = 1.;
	/* get colors (diffuse and specular for each face) */
	fscanf(fd,"%*s\n");
	for (iface = 0; iface < polytope->num_faces;iface++) {
		   fscanf(fd, "%lf %lf %lf\n", &O_DIFFUSE[iface][RED_IND],
		   &O_DIFFUSE[iface][GRN_IND], &O_DIFFUSE[iface][BLU_IND]);	
		   O_DIFFUSE[iface][RED_IND] = O_DIFFUSE[iface][RED_IND]/255.;
		   O_DIFFUSE[iface][GRN_IND] = O_DIFFUSE[iface][GRN_IND]/255.;
 		   O_DIFFUSE[iface][BLU_IND] = O_DIFFUSE[iface][BLU_IND]/255.;
	}
 	fscanf(fd,"%*s\n");
	for (iface = 0; iface < polytope->num_faces;iface++) {
		   fscanf(fd, "%lf %lf %lf\n", &O_SPECULAR[iface][RED_IND],
		     &O_SPECULAR[iface][GRN_IND], &O_SPECULAR[iface][BLU_IND]);
	 	   O_SPECULAR[iface][RED_IND] = O_SPECULAR[iface][RED_IND]/255.;
		   O_SPECULAR[iface][GRN_IND] = O_SPECULAR[iface][GRN_IND]/255.;
 		   O_SPECULAR[iface][BLU_IND] = O_SPECULAR[iface][BLU_IND]/255.;
	}
 
	/* set up edge table */
	if (first_time) {
          ET = (struct ET_element **) malloc((MAX_PIXELS + 1)*
                sizeof(struct ET_element *));
          /* set up active edge table */
          AET = (struct ET_element *) malloc(sizeof(struct ET_element));
          totpols = 0;
          for (iface = 0;iface < polytope->num_faces;iface++) totpols++;
          printf("TOTPOLS: %d\n", totpols);
          PT = (struct PT_struct *) malloc((totpols)*sizeof(struct PT_struct));
	}
	if (first_time) first_time = GL_FALSE;

}

/********************************************************/
/* FUNCTION: write_polytopes()				*/
/* PURPOSE:  this function writes all the information   */
/* 	     about the polytopes to an output file named */
/*	     "outpoly."					*/
/********************************************************/
static void
write_polytopes(tope_name)
	char *tope_name;
{

	GLint ifv, ivert, icell, iedge, iface;
	
	dfd = fopen(DEBUGFILE,"w");              
	
	fprintf(dfd,"\n ------ POLYTOPE %s ------- \n ", tope_name);
		
	fprintf(dfd,"\t (origin at: x = %lf y = %lf z = %lf w = %lf) \n \n",
			polytope->origin.p[X_COORD],
			polytope->origin.p[Y_COORD],
			polytope->origin.p[Z_COORD],
			polytope->origin.p[W_COORD]);

	for (ivert = 0; ivert < polytope->num_vertices;ivert++) {
			fprintf(dfd,
			  "\t vertex %d :  x = %lf y = %lf z = %lf w = %lf\n",
			  ivert,
			  polytope->vertex[ivert].p[X_COORD],
			  polytope->vertex[ivert].p[Y_COORD],
			  polytope->vertex[ivert].p[Z_COORD],
			  polytope->vertex[ivert].p[W_COORD]);
	}

	fprintf(dfd,"\n \n");
	for (iedge = 0; iedge < polytope->num_edges;iedge++) {
			fprintf(dfd,"\t edge %d : vertices %d and %d \n",
			  iedge,
			  polytope->edge[iedge].a,
			  polytope->edge[iedge].b);
	}

	fprintf(dfd,"\n \n");
	for (iface = 0; iface < polytope->num_faces;iface++) {
			fprintf(dfd,"\t face %d: \n",iface);
			for (ifv = 0; ifv <
			polytope->face[iface].num_vertices;ifv++) 
				fprintf(dfd,"\t \t vertex: %d\n",
			        polytope->face[iface].vertex[ifv]);
			for (ifv = 0; ifv <
			polytope->face[iface].num_vertices;ifv++) 
				fprintf(dfd,"\t \t edge: %d\n",
			        polytope->face[iface].edge[ifv]);
	}

 	/* print out cells to output file */
 	fprintf(dfd,"\n \n");
	for (icell = 0; icell < polytope->num_cells; icell++) {
                    fprintf(dfd, " cell %d: \n", icell);
		   for (iface = 0; 
		iface < polytope->cell[icell].num_faces; iface++) {
                        fprintf(dfd,"\t face %d: ", iface);
                        for (ivert = 0; ivert < 
		polytope->cell[icell].face[iface].num_vertices;
				 ivert++)
                        fprintf(dfd, "%d ",
		polytope->cell[icell].face[iface].vertex[ivert]);
                        fprintf(dfd,"\n");
		    }
	      	    fprintf(dfd,"\n");
        }

	fclose(dfd);
}

void
scale_to_screen_coords()
/********************************************************/
/* FUNCTION: convert_to_screen_coords()                 */
/* PURPOSE:  this function converts vertices expressed  */
/*           in world coordinates to screen coordinates */
/********************************************************/
{
        GLint iface, ivert;

#define V_sx                    (width_of_window/2)
#define V_sy                    (height_of_window/2)
#define V_sz                    ((f - n)/2)
#define V_cx                    (width_of_window/2)
#define V_cy                    (height_of_window/2)
#define V_cz                    ((n + f)/2)
     
	for (ivert = 0; ivert < polytope->num_vertices; ivert++) {
                  polytope->screen_vtx[ivert].p[X_COORD] =
                         (polytope->screen_vtx[ivert].p[X_COORD])
			 * V_sx + V_cx;
                  polytope->screen_vtx[ivert].p[Y_COORD] =
                        (polytope->screen_vtx[ivert].p[Y_COORD])
			 * V_sy + V_cy;
                  polytope->screen_vtx[ivert].p[Z_COORD] =
                        (polytope->screen_vtx[ivert].p[Z_COORD]) 
			* V_sz + V_cz;
		  polytope->screen_vtx[ivert] .p[W_COORD] = 0.0;
 	}
}

void
convert_to_ncube_coords()
/********************************************************/
/* FUNCTION: scale_to_screen_coords()                   */
/* PURPOSE:  this function converts vertices expressed  */
/*           in world coordinates to cube coordinates   */
/********************************************************/
{
        GLint iface, ivert;
     
	for (ivert = 0; 
		ivert < polytope->num_vertices; ivert++) {
                  (void) transf_vtx_to_screen(ivert);
	}
}
 
/**********************************************************/
/* FUNCTION: transf_vtx_to_screen()                       */
/* PURPOSE: this function transforms a given vertex by    */
/* the perspective matrix to get screen coordinates.      */
/**********************************************************/
void
transf_vtx_to_screen(ivtx)
	GLint ivtx;
{
        struct point tmp;
        GLdouble ctm[4][4];
        GLint i;

        glPushMatrix();
        glGetDoublev(GL_PROJECTION_MATRIX, ctm);


        tmp.p[X_COORD] = polytope->v3D[ivtx].p[X_COORD]*ctm[0][X_COORD] +
                        polytope->v3D[ivtx].p[Y_COORD]*ctm[1][X_COORD] +
                        polytope->v3D[ivtx].p[Z_COORD]*ctm[2][X_COORD] +
                        polytope->v3D[ivtx].p[W_COORD]*ctm[3][X_COORD];
        tmp.p[Y_COORD] = polytope->v3D[ivtx].p[X_COORD]*ctm[0][Y_COORD] +
                        polytope->v3D[ivtx].p[Y_COORD]*ctm[1][Y_COORD] +
                        polytope->v3D[ivtx].p[Z_COORD]*ctm[2][Y_COORD] +
                        polytope->v3D[ivtx].p[W_COORD]*ctm[3][Y_COORD];
        tmp.p[Z_COORD] = polytope->v3D[ivtx].p[X_COORD]*ctm[0][Z_COORD] +
                        polytope->v3D[ivtx].p[Y_COORD]*ctm[1][Z_COORD] +
                        polytope->v3D[ivtx].p[Z_COORD]*ctm[2][Z_COORD] +
                        polytope->v3D[ivtx].p[W_COORD]*ctm[3][Z_COORD];
        tmp.p[W_COORD] = polytope->v3D[ivtx].p[X_COORD]*ctm[0][W_COORD] +
                        polytope->v3D[ivtx].p[Y_COORD]*ctm[1][W_COORD] +
                        polytope->v3D[ivtx].p[Z_COORD]*ctm[2][W_COORD] +
                        polytope->v3D[ivtx].p[W_COORD]*ctm[3][W_COORD];
 	
        polytope->screen_vtx[ivtx].p[X_COORD] = 
		tmp.p[X_COORD]/fabs(tmp.p[W_COORD]);
        polytope->screen_vtx[ivtx].p[Y_COORD] = 
		tmp.p[Y_COORD]/fabs(tmp.p[W_COORD]);
        polytope->screen_vtx[ivtx].p[Z_COORD] =  
			tmp.p[Z_COORD]/fabs(tmp.p[W_COORD]);
        polytope->screen_vtx[ivtx].p[W_COORD] = tmp.p[W_COORD];

        glPopMatrix();
                
}       
 
/**********************************************************/
/* FUNCTION: compute_gouraud_4D_normals()                  */
/* PURPOSE: this function computes the volume normal of   */
/* a given vertex using the average volume normals of     */
/* each cell to which the vertex belongs.. (this is a 4D  */
/* analogy to averaging surface normal of polygonal faces */
/* (see Foley and van Dam , pp. 736 - 737.		  */
/**********************************************************/
void
compute_gouraud_4D_normals()
{
	GLdouble count = 0.0;
	GLint i, icell, iface, ivtx;
	
	for (i = 0; i < polytope->num_vertices; i++) {

	   polytope->gouraud_norm[i].p[X_COORD] = 0.0;
	   polytope->gouraud_norm[i].p[Y_COORD] = 0.0;
	   polytope->gouraud_norm[i].p[Z_COORD] = 0.0;
	   polytope->gouraud_norm[i].p[W_COORD] = 0.0;

	   for (icell = 0; icell< polytope->num_cells; icell++) {

             if (cell_contains_vtx(icell, i)) {

		   /* average these normals to get Gouraud shading */
                   polytope->gouraud_norm[i].p[X_COORD] =
	              polytope->gouraud_norm[i].p[X_COORD] 
                      + polytope->cell[icell].volume_normal.p[X_COORD];
                   polytope->gouraud_norm[i].p[Y_COORD] =      
		      polytope->gouraud_norm[i].p[Y_COORD] 
                      + polytope->cell[icell].volume_normal.p[Y_COORD];
                   polytope->gouraud_norm[i].p[Z_COORD] =     
		      polytope->gouraud_norm[i].p[Z_COORD] 
                      + polytope->cell[icell].volume_normal.p[Z_COORD];
		   polytope->gouraud_norm[i].p[W_COORD] =      
		      polytope->gouraud_norm[i].p[W_COORD] 
                      + polytope->cell[icell].volume_normal.p[W_COORD];
		   count = count + 1.0;
              }
           }
	
	   polytope->gouraud_norm[i].p[X_COORD] = 
		polytope->gouraud_norm[i].p[X_COORD]/count;
           polytope->gouraud_norm[i].p[Y_COORD] = 
		polytope->gouraud_norm[i].p[Y_COORD]/count;
           polytope->gouraud_norm[i].p[Z_COORD] = 
		polytope->gouraud_norm[i].p[Z_COORD]/count;
           polytope->gouraud_norm[i].p[W_COORD] = 
		polytope->gouraud_norm[i].p[W_COORD]/count;
	     normalize(polytope->gouraud_norm[i].p);

	}

	for (iface = 0;iface < polytope->num_faces; iface++) {

		PT[iface].gouraud_norm.p[X_COORD] = 0.0;
		PT[iface].gouraud_norm.p[Y_COORD] = 0.0;
		PT[iface].gouraud_norm.p[Z_COORD] = 0.0;
		PT[iface].gouraud_norm.p[W_COORD] = 0.0;

		for (ivtx = 0;ivtx < polytope->face[iface].num_vertices;
			ivtx++) {
		    PT[iface].gouraud_norm.p[X_COORD] = 
			polytope->gouraud_norm[ivtx].p[X_COORD] + 
			PT[iface].gouraud_norm.p[X_COORD];
  		    PT[iface].gouraud_norm.p[Y_COORD] = 
			polytope->gouraud_norm[ivtx].p[Y_COORD] + 
			PT[iface].gouraud_norm.p[Y_COORD];
  		    PT[iface].gouraud_norm.p[Z_COORD] = 
			polytope->gouraud_norm[ivtx].p[Z_COORD] + 
			PT[iface].gouraud_norm.p[Z_COORD];
  		    PT[iface].gouraud_norm.p[W_COORD] = 
			polytope->gouraud_norm[ivtx].p[W_COORD] + 
			PT[iface].gouraud_norm.p[W_COORD];
		   count = count + 1.0;
                }

               PT[iface].gouraud_norm.p[X_COORD] = 
		 PT[iface].gouraud_norm.p[X_COORD]/count;
                PT[iface].gouraud_norm.p[Y_COORD] = 
		 PT[iface].gouraud_norm.p[Y_COORD]/count;
                PT[iface].gouraud_norm.p[Z_COORD] = 
		 PT[iface].gouraud_norm.p[Z_COORD]/count;
                PT[iface].gouraud_norm.p[W_COORD] = 
		 PT[iface].gouraud_norm.p[W_COORD]/count;
	        normalize(PT[iface].gouraud_norm.p);
        }


}

GLboolean
cell_contains_vtx(icell, ivtx)
	int icell, ivtx;
{
	GLint iface, jvtx;
	for (iface = 0;iface < polytope->cell[icell].num_faces; iface++) 
	    for (jvtx = 0;
		jvtx < polytope->cell[icell].face[iface].num_vertices;
		jvtx++)
		if (ivtx == polytope->cell[icell].face[iface].vertex[jvtx])
			return (GL_TRUE);
	return (GL_FALSE);
}

/**********************************************************/
/* FUNCTION: compute_gouraud_3D_normal()                  */
/* PURPOSE: this function computes the surface normal of  */
/* a given vertex using the average surface normals of    */
/* each face to which the vertex belongs..		  */
/* (see Foley and van Dam , pp. 736 - 737.		  */
/**********************************************************/
void
compute_gouraud_3D_normals()
{
	GLdouble count = 0.0;
	GLint ivtx, iface, i;

	 for (i = 0; i < polytope->num_vertices; i++) {

	    polytope->gouraud_norm[i].p[X_COORD] = 0.0;
	    polytope->gouraud_norm[i].p[Y_COORD] = 0.0;
	    polytope->gouraud_norm[i].p[Z_COORD] = 0.0;
	    polytope->gouraud_norm[i].p[W_COORD] = 0.0;

	    for (iface = 0; iface< polytope->num_faces; iface++) {

             if (face_contains_vtx(iface, i)) {

		   /* average these normals to get Gouraud shading */
                   polytope->gouraud_norm[i].p[X_COORD] = 
		      polytope->gouraud_norm[i].p[X_COORD] 
                      + polytope->face[iface].surface_normal.p[X_COORD];
                   polytope->gouraud_norm[i].p[Y_COORD] = 
		      polytope->gouraud_norm[i].p[Y_COORD] 
                      + polytope->face[iface].surface_normal.p[Y_COORD];
                   polytope->gouraud_norm[i].p[Z_COORD] = 
		      polytope->gouraud_norm[i].p[Z_COORD] 
                      + polytope->face[iface].surface_normal.p[Z_COORD];
		   polytope->gouraud_norm[i].p[W_COORD] = 
		      polytope->gouraud_norm[i].p[W_COORD] 
                      + polytope->face[iface].surface_normal.p[W_COORD];
		   count = count + 1.0;
              }
            }
	
	    polytope->gouraud_norm[i].p[X_COORD] = 
                polytope->gouraud_norm[i].p[X_COORD]/count;
            polytope->gouraud_norm[i].p[Y_COORD] = 
                polytope->gouraud_norm[i].p[Y_COORD]/count;
            polytope->gouraud_norm[i].p[Z_COORD] = 
                polytope->gouraud_norm[i].p[Z_COORD]/count;
            polytope->gouraud_norm[i].p[W_COORD] = 
                polytope->gouraud_norm[i].p[W_COORD]/count;
	    normalize(polytope->gouraud_norm[i].p);

	}

	for (iface = 0;iface < polytope->num_faces; iface++) {

		PT[iface].gouraud_norm.p[X_COORD] = 0.0;
		PT[iface].gouraud_norm.p[Y_COORD] = 0.0;
		PT[iface].gouraud_norm.p[Z_COORD] = 0.0;
		PT[iface].gouraud_norm.p[W_COORD] = 0.0;

		for (ivtx = 0;ivtx < polytope->face[iface].num_vertices;
			ivtx++) {
		    PT[iface].gouraud_norm.p[X_COORD] = 
			polytope->gouraud_norm[ivtx].p[X_COORD] + 
			PT[iface].gouraud_norm.p[X_COORD];
  		    PT[iface].gouraud_norm.p[Y_COORD] = 
			polytope->gouraud_norm[ivtx].p[Y_COORD] + 
			PT[iface].gouraud_norm.p[Y_COORD];
  		    PT[iface].gouraud_norm.p[Z_COORD] = 
			polytope->gouraud_norm[ivtx].p[Z_COORD] + 
			PT[iface].gouraud_norm.p[Z_COORD];
  		    PT[iface].gouraud_norm.p[W_COORD] = 
			polytope->gouraud_norm[ivtx].p[W_COORD] + 
			PT[iface].gouraud_norm.p[W_COORD];
		   count = count + 1.0;
                }

               PT[iface].gouraud_norm.p[X_COORD] = 
		 PT[iface].gouraud_norm.p[X_COORD]/count;
                PT[iface].gouraud_norm.p[Y_COORD] = 
		 PT[iface].gouraud_norm.p[Y_COORD]/count;
                PT[iface].gouraud_norm.p[Z_COORD] = 
		 PT[iface].gouraud_norm.p[Z_COORD]/count;
                PT[iface].gouraud_norm.p[W_COORD] = 
		 PT[iface].gouraud_norm.p[W_COORD]/count;
	        normalize(PT[iface].gouraud_norm.p);
        }

}

static GLboolean
face_contains_vtx(iface, ivtx)
	GLint iface, ivtx;
{
	GLint  jvtx;

       for (jvtx = 0;
	  jvtx < polytope->face[iface].num_vertices;jvtx++)
	  if (ivtx == polytope->face[iface].vertex[jvtx])
		return (GL_TRUE);
	return (GL_FALSE);
}

/**********************************************************/
/* FUNCTION: compute_volume_normals()                     */
/* PURPOSE: this function computes the volume normal of   */
/* a all cells of the polytope. 		          */
/**********************************************************/

void
compute_volume_normals()
{
        GLint icell, vert1, vert2, vert3, vert4;
        struct point vol_normal, r_start, r_end, s_start, s_end, 
		t_start, t_end, r, s, t;
        GLdouble r1,r2,r3,r4,s1,s2,s3,s4,t1,t2,t3,t4;
 
       	dfd = fopen(DEBUGFILE, "a");
	for (icell = 0; icell < polytope->num_cells; icell++) {
 
 
        /* compute volume normal of this cell */
        vert1 = polytope->cell[icell].face[0].vertex[0];
        vert2 = polytope->cell[icell].face[0].vertex[1];
        vert3 = polytope->cell[icell].face[0].vertex[2];
        vert4 = polytope->cell[icell].face[1].vertex[0];
 
        r_start =  polytope->rot_vertex[vert2];
        r_end = polytope->rot_vertex[vert1];
        s_start = r_start;
        s_end = polytope->rot_vertex[vert3];
        t_start = r_start;
        t_end = polytope->rot_vertex[vert4];
 
        r = vector_minus(r_end, r_start);
        s = vector_minus(s_end, s_start);
        t = vector_minus(t_end, t_start);
                                               
        r1=r.p[X_COORD];r2=r.p[Y_COORD];r3=r.p[Z_COORD];r4=r.p[W_COORD];
        s1=s.p[X_COORD];s2=s.p[Y_COORD];s3=s.p[Z_COORD];s4=s.p[W_COORD];
        t1=t.p[X_COORD];t2=t.p[Y_COORD];t3=t.p[Z_COORD];t4=t.p[W_COORD];

        /* print r, s and t */
	fprintf(dfd, "* * * CELL %d VOLUME NORMAL CALCULATIONS * * *\n", icell);
        (void) print_point("R",r);
        (void) print_point("S",s);
        (void) print_point("T",t);
	 
        /* compute intermediate expressions */
 	vol_normal.p[X_COORD] = det(r2, r3, r4, s2, s3, s4, t2, t3, t4);
	vol_normal.p[Y_COORD] = -1.*det(r1, r3, r4, s1, s3, s4, t1, t3, t4);
	vol_normal.p[Z_COORD] = det(r1, r2, r4, s1, s2, s4, t1, t2, t4);
	vol_normal.p[W_COORD] = -1*det(r1, r2, r3, s1, s2, s3, t1, t2, t3);
                     
        (void) normalize(vol_normal.p);
        polytope->cell[icell].volume_normal.p[X_COORD] = 
			vol_normal.p[X_COORD];
        polytope->cell[icell].volume_normal.p[Y_COORD] = 
			vol_normal.p[Y_COORD];
        polytope->cell[icell].volume_normal.p[Z_COORD] = 
			vol_normal.p[Z_COORD];
        polytope->cell[icell].volume_normal.p[W_COORD] = 
			vol_normal.p[W_COORD];
 	print_point("volume_normal",polytope->cell[icell].volume_normal);


       }

       fclose(dfd);


}

/**********************************************************/
/* FUNCTION: compute_surface_normals()                    */
/* PURPOSE: this function computes the surface normal of  */
/* a all faces of the polytope. 		          */
/**********************************************************/

void
compute_surface_normals()
{
        GLint iface, vert1, vert2, vert3, vert4;
        struct point surf_normal, r_start, r_end, s_start, s_end, 
		t_start, t_end, r, s, t;
        double r1,r2,r3,r4,s1,s2,s3,s4; 
       	dfd = fopen(DEBUGFILE, "a");
	for (iface = 0; iface < polytope->num_faces; iface++) {
 
 
        /* compute volume normal of this face */
        vert1 = polytope->face[iface].vertex[0];
        vert2 = polytope->face[iface].vertex[1];
        vert3 = polytope->face[iface].vertex[2];
        
        r_start =  polytope->rot_vertex[vert2];
        r_end = polytope->rot_vertex[vert1];
        s_start = r_start;
        s_end = polytope->rot_vertex[vert3];
       
        r = vector_minus(r_end, r_start);
        s = vector_minus(s_end, s_start);
                                                     
        r1=r.p[X_COORD];r2=r.p[Y_COORD];r3=r.p[Z_COORD];r4=r.p[W_COORD];
        s1=s.p[X_COORD];s2=s.p[Y_COORD];s3=s.p[Z_COORD];s4=s.p[W_COORD];
      
        /* print r, s and t */
	fprintf(dfd, "* * * FACE %d SURFACE NORMAL CALCULATIONS * * *\n", 
		iface);
        (void) print_point("R",r);
        (void) print_point("S",s);
    	 
        /* compute intermediate expressions */
 	surf_normal.p[X_COORD] = r2*s3 - s2*r3;
	surf_normal.p[Y_COORD] = r3*s1 - s3*r1;
	surf_normal.p[Z_COORD] = r1*s2 - r2*s1;
	surf_normal.p[W_COORD] = 0.0;
                     
        (void) normalize(surf_normal.p);
        polytope->face[iface].surface_normal.p[X_COORD] = 
			surf_normal.p[X_COORD];
        polytope->face[iface].surface_normal.p[Y_COORD] = 
			surf_normal.p[Y_COORD];
        polytope->face[iface].surface_normal.p[Z_COORD] = 
			surf_normal.p[Z_COORD];
        polytope->face[iface].surface_normal.p[W_COORD] = 
			surf_normal.p[W_COORD];
 	print_point("surface_normal",polytope->face[iface].surface_normal);

       }

       fclose(dfd);


}

struct point
vector_minus(a, b)
	struct point a, b;
{
        static struct point n;

         /* subtract components of these vectors */
         n.p[X_COORD] = a.p[X_COORD] - b.p[X_COORD];
         n.p[Y_COORD] = a.p[Y_COORD] - b.p[Y_COORD];
         n.p[Z_COORD] = a.p[Z_COORD] - b.p[Z_COORD];
         n.p[W_COORD] = a.p[W_COORD] - b.p[W_COORD];

         return (n);
}

void
normalize(a)
	GLdouble *a;
{
	GLdouble length = 0.0;
	GLint i;
	
	length = sqrt(a[X_COORD]*a[X_COORD] + a[Y_COORD]*a[Y_COORD]
		+ a[Z_COORD]*a[Z_COORD] + a[W_COORD]*a[W_COORD]);

	if (length > EPSILON) {
	    a[X_COORD] = a[X_COORD]/length;
  	    a[Y_COORD] = a[Y_COORD]/length;
   	    a[Z_COORD] = a[Z_COORD]/length;
   	    a[W_COORD] = a[W_COORD]/length;	
	    return;
	}
	else {

	   if (length < -1.* EPSILON)
	     printf(" \n ERROR: vector won't normalize\n"
			" %lf %lf %lf %lf l:%lf\n",
	      a[X_COORD], a[Y_COORD], a[Z_COORD], a[W_COORD], length);
		
	}
}

void
print_point(title, a)
	char *title;
	struct point a;
{
        fprintf(dfd, "%s: X = %10.6lf Y = %10.6f "
                        "Z = %10.6lf W = %10.6lf\n",
                title, a.p[X_COORD], a.p[Y_COORD],
                        a.p[Z_COORD], a.p[W_COORD]);
}

extern double
det(a, b, c, d, e, f, g, h, i)
	GLdouble a, b, c, d, e, f, g, h, i;
{
	GLdouble ret_val;

	ret_val = a*(e*i - f*h) - b*(d*i - f*g) + c*(d*h - e*g);
	return (ret_val);
}

extern void
mark_time(string)
        char *string;
{
        long clock;
        char *ascii_time;

        clock = time((long *) 0);
        ascii_time = ctime(&clock);

        fprintf(dfd, "%s\n\t%ld = %s", string, clock, ascii_time);
}

extern GLboolean
cross()
{
	if (polytope->num_vertices == 8) return GL_TRUE;
	else return GL_FALSE;
}
